
//load "~/magma/robin/phi345.m";
//
Attach("phi345.m");

import "phi345.m" : phi3;
import "phi345.m" : phi4nf;
import "phi345.m" : phi5nf;


////////////////////////////////////////////////////////////////////////
//                                                                        
//      Computation of canonical heights on genus 2 jacobians
//      using      division polynomials
//                                                                                                  
//      Jan Steffen Mueller (February 2013),
//      based on the paper "Canonical heights and division polynomials"
//      by J.S. Mueller and R. de Jong.
//                                                                        
////////////////////////////////////////////////////////////////////////


// todo: 
// - estimate precision needed
// - create a memory-efficient version -> store only needed division values.
// - more comments
// - checks
// - tests


function variable_Di(f, i, coefs, vars) 
// Given an element f of vars = [p11,...,p222], an integer i and a set of coefficients specifying
// a genus 2 curve, return the derivation D_i(f) as in Kanayama '05, Theorem 9.

    f4, f3, f2, f1, f0 := Explode(coefs);
    case f:    
        when vars[1]:
            return i eq 1 select vars[4] else vars[5];
        when vars[2]:
            return i eq 1 select vars[5] else vars[6];
        when vars[3]:
            return i eq 1 select vars[6] else vars[7];
        when vars[4]:
            return i eq 1 select 6 * vars[1]^2 + 4 * f2 * vars[1] + 4 * f1 * vars[2] - 12 * f0 *
vars[3] - 8 * f4 * f0 + 2 * f3 * f1 else 6 * vars[1] * vars[2] + 4 * f2 * vars[2] - 2 * f1 *
vars[3] - 4 * f0;
        when vars[5]:
            return i eq 1 select 6 * vars[1] * vars[2] + 4 * f2 * vars[2] - 2 * f1 *
            vars[3] - 4 * f0
else 2 * vars[1] * vars[3] + 4 * vars[2]^2 + 2 * f3 * vars[2];
        when vars[6]:
            return i eq 1 select 2 * vars[1] * vars[3] + 4 * vars[2]^2 + 2 * f3 * vars[2]
else 6 * vars[2] *
vars[3] - 2 * vars[1] + 4 * f4 * vars[2];
        when vars[7]:
            return i eq 1 select 6 * vars[2] * vars[3] - 2 * vars[1] + 4 * f4 * vars[2]
else 6 * vars[3]^2
+ 4 * vars[2] + 4 * f4 * vars[3] + 2 * f3;
    end case;
end function;


function monomial_Di(f, i, j, coefs, vars)
// Given a monomial f in vars = [p11,...,p222], an integer i and a set of coefficients specifying
// a genus 2 curve, return the derivation D_i(f) as in Kanayama '05, Theorem 9.

    if IsOne(f) then
        return 0;
    end if;
    deg := Degree(f, j); 
    if IsZero(deg) then
        return monomial_Di(f, i, j + 1, coefs, vars);
    else
        pj := vars[j];
        g := f div pj^deg; 
        return pj^deg * monomial_Di(g, i, j + 1, coefs, vars) + deg * pj^(deg - 1) * g * variable_Di(pj, i,
            coefs, vars); 
    end if;
end function;


function Di(f, i, coefs, vars)
// Given a polynomial f in vars = [p11,...,p222], an integer i and a set of coefficients specifying
// a genus 2 curve, return the derivation D_i(f) as in Kanayama '05, Theorem 9.

    C, M := CoefficientsAndMonomials(f);
    return &+[C[j] * monomial_Di(M[j], i, 1, coefs, vars) : j in [1..#M]];
end function;


function initial_division_polynomials(coeffs)
// Compute the first 5 division polynomials associated to the Jacobian of the genus 2 curve
// specified by the coefficients coefs = [f4, f3, f2, f1, f0]

    R := FieldOfFractions(Universe(coeffs));
    _<p11, p12, p22, p111, p112, p122, p222> := PolynomialRing(R, 7);
    L := [One(R), p12 * p122 - p22 * p112 - p111];
    p := [p11, p12, p22, p111, p112, p122, p222];
    Append(~L, phi3(p, coeffs));
    Append(~L, phi4nf(p, coeffs));
    Append(~L, phi5nf(p, coeffs));
    return L;
end function;


function division_values(P, fs, bd)
//    P: point on a Jacobian surface, not in the support of the theta divisor
//    C := Curve(Parent(P));
//    error if Degree(P[1]) lt Genus(C), "P lies in the theta divisor";
//    fs := Coefficients(HyperellipticPolynomials(C));

    coefs := Reverse(Prune(fs));
    phi := initial_division_polynomials(coefs);
    PR := Parent(phi[2]);
    vars := [PR.i : i in [1..7]];
    g := 2;  // todo: generalize!
    pgis := [-Coefficient(P[1], g - i -1) : i in [0..g - 1]]; 
    pggis := [2 * Coefficient(P[2], g - i - 1) : i in [0..g - 1]]; 
    // now add p_11, p_111, p_112
    Append(~pgis, -pgis[1]^3 - pgis[1] * pgis[2] - fs[5] * pgis[1]^2 - fs[4] *
    pgis[1] - fs[3] + pggis[1]^2 / 4); //p11
    Append(~pggis, pgis[2] * pggis[1] - pgis[1] * pggis[2]); // p112
    Append(~pggis, 2 * (pgis[1] + fs[5]) * pggis[3] - (pgis[2] + fs[4]) * pggis[2]
    - pgis[3] * pggis[1]); // p111
    // The orders are [p22, p12, p11] and [p222, p122, p112, p111].

    ps := Reverse(pgis) cat Reverse(pggis);
//    "Computing phi_1(P),...,phi_5(P)";
    L := [Evaluate(p, ps) : p in phi];
    if bd le 5 then return L; end if;
    error if L[2] eq 0, "2P lies in the theta divisor";
    hms := [];
    for n := 2 to 5 do
     //   printf "Computing h^(i)_%o\n",n;
        derivs_n1 := Evaluate(Di(phi[n], 1, coefs, vars), ps);
        d_n2 := Di(phi[n], 2, coefs, vars);
        derivs_n2 := Evaluate(d_n2, ps); 
        dderivs_n1 := Evaluate(Di(d_n2, 1, coefs, vars), ps);
        dderivs_n2 := Evaluate(Di(d_n2, 2, coefs, vars), ps); 
        Append(~hms, [derivs_n1 * derivs_n2 - L[n] * dderivs_n1,
                    derivs_n2^2 - L[n] * dderivs_n2]);
    end for;
    Append(~L, (L[2]^2 * L[3] * L[5] - L[4]^2 * L[3] * L[1] + (hms[1, 1] * hms[3, 2] -
    hms[1, 2] * hms[3, 1]) / 64) / L[2]);  // phi6
    Append(~L, L[3]^3 * L[5] - L[4]^3 * L[2] + (hms[2, 1] * hms[3, 2] - hms[2, 2] * hms[3,
    1]) / 144); //  phi7
    Append(~L, (L[3]^2 * L[4] * L[6] - L[5]^2 * L[4] * L[2] + (hms[2, 1] * hms[4, 2] -
    hms[2, 2] * hms[4, 1]) / 225) / L[2]);  // phi8
    if bd le 8 then return L; end if;
    quotient := L[5] - L[4] * L[2]^3 + L[3]^3;
    error if quotient eq 0, "quotient vanishes at P";

    m := 4; 
    //"starting recursion";
    while 2 * m le bd do
        if m gt 4 then
//            printf "computing phi_%o(P)\n", 2*m;
            Append(~L, (L[m + 4] * L[m + 1] * L[m - 2] * L[m - 3] - L[m + 4] * L[m] * L[m - 2]^2 * L[2]^2 + L[m + 4] * L[m - 1]^2 * L[m - 2] * L[3] - L[m + 3] * L[m + 2] * L[m - 1] * L[m - 4] + L[m + 3] * L[m] * L[m - 1] * L[m - 2] * L[3]^2 - L[m + 3] * L[m - 1]^3 * L[4] * L[2] + L[m + 2]^2 * L[m] * L[m - 4] * L[2]^2 - L[m + 2] * L[m + 1] * L[m] * L[m - 3] * L[3]^2 + L[m + 2] * L[m] * L[m - 1]^2 * L[5] - L[m + 2] * L[m + 1]^2 * L[m - 4] * L[3] + L[m + 1]^3 * L[m - 3] * L[4] * L[2] - L[m + 1]^2 * L[m] * L[m - 2] * L[5]) / L[2] / quotient);
        end if;
        if 2 * m ne bd then 
//            printf "computing phi_%o(P)\n", 2*m+1;
            Append(~L, (L[m + 4] * L[m + 2] * L[m - 2]^2 - L[m + 4] * L[m + 1] * L[m - 1] * L[m - 2] * L[2]^2 + L[m + 4] * L[m]^2 * L[m - 2] * L[3] - L[m + 3]^2 * L[m - 1] * L[m - 3] + L[m + 3] * L[m + 1] * L[m - 1]^2 * L[3]^2 - L[m + 3] * L[m]^2 * L[m - 1] * L[4] * L[2] + L[m + 3] * L[m + 2] * L[m] * L[m - 3] * L[2]^2 - L[m + 2]^2 * L[m] * L[m - 2] * L[3]^2 + L[m + 2] * L[m]^3 * L[5] - L[m + 3] * L[m + 1]^2 * L[m - 3] * L[3] + L[m + 2] * L[m + 1]^2 * L[m - 2] * L[4] * L[2] - L[m + 1]^3 * L[m - 1] * L[5]) / quotient);
        end if;
        m +:= 1;
    end while;
    return L; 
end function;


function division_value(P, fs, bd)
   return division_values(P, fs, bd)[bd]; 
end function;


intrinsic LocalHeightUchida(P::JacHypPt, v::RngIntElt: Iterations := 0, Precision := 0)
    -> FldReElt
{Computes the local height at v of a point P on the Jacobian of a genus 2 curve over the 
    rationals given by a monic odd degree model, where v is a place. This uses Theorem 
    2.1 of the paper Canonical heights and division polynomials by J.S. Mueller and R. de Jong.}

    C := Curve(Parent(P));
    require Genus(C) eq 2:
            "The genus of the curve must be 2.";    
    f, h := HyperellipticPolynomials(C);
    require IsZero(h):
            "The curve must be given by a simplified model.";
    require IsOdd(Degree(f)):
            "The curve must be given by a monic odd degree model.";
    require ISA(Type(BaseRing(C)), FldRat):
            "The base field must be the rational field.";

    if IsZero(Iterations) then
        Iterations := 1000;
    end if;
    if IsZero(Precision) then
        Precision := 50;
    end if;
    fs := Coefficients(HyperellipticPolynomials(Curve(Parent(P))));
    if v ne 0 then
        // Work over Qp
        Qp := pAdicField(v, Precision); 
        Pv := [ChangeRing(g, Qp) : g in [P[1], P[2]]];
        // Compute phi_n(Pv), n = Iterations.
        phi_v := division_value(Pv, fs, Iterations);
        return -Valuation(phi_v) * Log(v) / Iterations^2;
    else
        // Work over R
        RF := RealField(Precision); 
        Pv := [ChangeRing(g, RF) : g in [P[1], P[2]]];
        // Compute phi_n(Pv), n = Iterations.
        phi_v := division_value(Pv, fs, Iterations);
        return Log(Abs(phi_v)) / Iterations^2;
    end if;
end intrinsic;



intrinsic Height1(P::JacHypPt : Iterations := 0, Precision := 0) -> FldReElt
{Computes the canonical height of a point P on the Jacobian 
    of a genus 2 curve over the rationals given by a monic odd degree model. This uses Theorem 
    2.4 of the paper Canonical heights and division polynomials by J.S. Mueller and R. de Jong.
    Note that this is half the height computed by Height(P).}

    C := Curve(Parent(P));
    require Genus(C) eq 2:
            "The genus of the curve must be 2.";    
    f, h := HyperellipticPolynomials(C);
    require IsZero(h):
            "The curve must be given by a simplified model.";
    require IsOdd(Degree(f)):
            "The curve must be given by a monic odd degree model.";
    require ISA(Type(BaseRing(C)), FldRat):
            "The base field must be the rational field.";

    if Order(P) eq 2 then
        return 0;
    end if;
    if IsZero(Iterations) then
        Iterations := 1000;
    end if;
    if IsZero(Precision) then
        Precision := 50;
    end if;

    // Find the primes that might contribute.
    disc := Integers() ! Discriminant(C);
    S := [0, 2] cat [p: p in PrimeDivisors(disc) | Valuation(disc, p) gt 1 and p ne 2];
    S cat:= [p : p in PrimeDivisors(LCM([Denominator(c) : c in Coefficients(P[1])])) | 
                                                          p notin  S];
    // Compute the local heights.
    local_heights := [LocalHeightUchida(P, v : Iterations := Iterations, Precision := 
                                               Precision) : v in S];
    return &+local_heights, [<S[i], local_heights[i]> : i in [1..#S]];
end intrinsic;


intrinsic Height2(P::JacHypPt : Iterations := 0) -> FldReElt
{Computes the canonical height of a point P with integral Mumford representation on the Jacobian 
    of a genus 2 curve over the rationals given by a monic odd degree model. This uses Corollary 
    8.3 of the paper Canonical heights and division polynomials by J.S. Mueller and R. de Jong.
    Note that this is half the height computed by Height(P).}

    C := Curve(Parent(P));
    require Genus(C) eq 2:
            "The genus of the curve must be 2.";    
    f, h := HyperellipticPolynomials(C);
    require IsZero(h):
            "The curve must be given by a simplified model.";
    require IsOdd(Degree(f)):
            "The curve must be given by a monic odd degree model.";
    require ISA(Type(BaseRing(C)), FldRat):
            "The base field must be the rational field.";
    require &and[IsIntegral(c): c in Coefficients(P[1]) cat Coefficients(P[2])]:
            "The Mumford representation of P must have integral coefficients.";
    
    if Order(P) eq 2 then
        return 0;
    end if;
    if GCD(P[1], P[2]) ne 1 then
        // The support of P contains a Weierstrass point.
    return Height2(2 * P : Iterations := Iterations) / 4;
    end if;
    if IsZero(Iterations) then
        Iterations := 500;
    end if;

    fs := Coefficients(f);
    n := Iterations - 2;
    Z := Integers();
    // Compute division values
    E := division_values(P, fs, Iterations);
    return Log(Abs(Z ! E[n] / GCD(ChangeUniverse([E[n], E[n + 1], E[n + 2]], Z)))) / n^2;
end intrinsic;


